import logging
import re
from lsprotocol.types import Position, Range
from typing import List, Optional

from pygls.workspace import Document
from vyper.ast import EnumDef, FunctionDef, VyperNode
from vyper_lsp.ast import AST
from vyper_lsp.utils import get_expression_at_cursor, get_word_at_cursor

ENUM_VARIANT_PATTERN = re.compile(r"([a-zA-Z_][a-zA-Z0-9_]*)\.([a-zA-Z_][a-zA-Z0-9_]*)")

logger = logging.getLogger("vyper-lsp")


# this class should abstract away all the AST stuff
# and just provide a simple interface for navigation
#
# the navigator should mainly return Ranges
class ASTNavigator:
    def __init__(self, ast: AST):
        self.ast = ast

    def _create_range_from_node(self, node: VyperNode) -> Range:
        return Range(
            start=Position(line=node.lineno - 1, character=node.col_offset),
            end=Position(line=node.end_lineno - 1, character=node.end_col_offset),
        )

    def _find_state_variable_declaration(self, word: str) -> Optional[Range]:
        node = self.ast.find_state_variable_declaration_node_for_name(word)
        if node:
            return self._create_range_from_node(node)

    def _find_variable_declaration_under_node(
        self, node: VyperNode, symbol: str
    ) -> Optional[Range]:
        decl_node = AST.from_node(node).find_node_declaring_symbol(symbol)
        if decl_node:
            return self._create_range_from_node(decl_node)

    def _find_function_declaration(self, word: str) -> Optional[Range]:
        node = self.ast.find_function_declaration_node_for_name(word)
        if node:
            return self._create_range_from_node(node)

    def find_type_declaration(self, word: str) -> Optional[Range]:
        node = self.ast.find_type_declaration_node_for_name(word)
        if node:
            return self._create_range_from_node(node)

    def _is_state_var_decl(self, line, word):
        is_top_level = not line[0].isspace()
        is_state_variable = word in self.ast.get_state_variables()
        return is_top_level and is_state_variable

    def _is_constant_decl(self, line, word):
        is_constant = "constant(" in line
        return is_constant and self._is_state_var_decl(line, word)

    def _is_internal_fn(self, line, word, expression):
        is_def = line.startswith("def")
        is_internal_call = expression.startswith("self.")
        is_internal_fn = word in self.ast.get_internal_functions()
        return is_def and (is_internal_call or is_internal_fn)

    def find_references(self, doc: Document, pos: Position) -> List[Range]:
        if self.ast.ast_data is None:
            return []
        references = []

        og_line = doc.lines[pos.line]
        word = get_word_at_cursor(og_line, pos.character)
        expression = get_expression_at_cursor(og_line, pos.character)

        top_level_node = self.ast.find_top_level_node_at_pos(pos)

        refs = []

        if word in self.ast.get_enums():
            # find all references to this type
            refs = self.ast.find_nodes_referencing_enum(word)
        elif word in self.ast.get_structs() or word in self.ast.get_events():
            refs = self.ast.find_nodes_referencing_struct(word)
        elif self._is_internal_fn(og_line, word, expression):
            refs = self.ast.find_nodes_referencing_internal_function(word)
        elif self._is_constant_decl(og_line, word):
            refs = self.ast.find_nodes_referencing_constant(word)
        elif self._is_state_var_decl(og_line, word):
            refs = self.ast.find_nodes_referencing_state_variable(word)
        elif isinstance(top_level_node, EnumDef):
            # find all references to this enum variant
            refs = self.ast.find_nodes_referencing_enum_variant(
                top_level_node.name, word
            )
        elif isinstance(top_level_node, FunctionDef):
            refs = AST.from_node(top_level_node).find_nodes_referencing_symbol(word)

        for ref in refs:
            range = self._create_range_from_node(ref)
            references.append(range)
        return references

    def _match_enum_variant(self, full_word: str) -> Optional[re.Match]:
        match = ENUM_VARIANT_PATTERN.match(full_word)
        if (
            match
            and match.group(1) in self.ast.get_enums()
            and match.group(2) in self.ast.get_enum_variants(match.group(1))
        ):
            return match
        return None

    def find_declaration(self, document: Document, pos: Position) -> Optional[Range]:
        if self.ast.ast_data is None:
            return None

        line_content = document.lines[pos.line]
        word = get_word_at_cursor(line_content, pos.character)
        full_word = get_expression_at_cursor(line_content, pos.character)
        top_level_node = self.ast.find_top_level_node_at_pos(pos)

        # Determine the type of declaration and find it
        if full_word.startswith("self."):
            if "(" in full_word:
                return self._find_function_declaration(word)
            else:
                return self._find_state_variable_declaration(word)
        elif word in self.ast.get_user_defined_types():
            return self.find_type_declaration(word)
        elif word in self.ast.get_events():
            return self.find_type_declaration(word)
        elif word in self.ast.get_constants():
            return self._find_state_variable_declaration(word)
        elif isinstance(top_level_node, FunctionDef):
            range = self._find_variable_declaration_under_node(top_level_node, word)
            if range:
                return range
            else:
                match = self._match_enum_variant(full_word)
                if match:
                    return self.find_type_declaration(match.group(1))

    def find_implementation(self, document: Document, pos: Position) -> Optional[Range]:
        og_line = document.lines[pos.line]
        word = get_word_at_cursor(og_line, pos.character)
        expression = get_expression_at_cursor(og_line, pos.character)

        if "(" not in expression:
            return None

        if expression.startswith("self."):
            return self._find_function_declaration(word)
        elif og_line[0].isspace() and og_line.strip().startswith("def"):
            # only lookup external fns if we're in an interface def
            return self._find_function_declaration(word)
        else:
            return None



import logging

logger = logging.getLogger("vyper-lsp")
logger.setLevel(logging.INFO)

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
)


class LanguageServerLogHandler(logging.Handler):
    def __init__(self, ls):
        super().__init__()
        self.ls = ls

    def emit(self, record):
        log_entry = self.format(record)
        if self.ls:
            self.ls.show_message_log(log_entry)



import logging
import re
from typing import List, Optional
import warnings
from packaging.version import Version
from lsprotocol.types import (
    Diagnostic,
    ParameterInformation,
    Position,
    Range,
    SignatureHelp,
    SignatureInformation,
)
from pygls.workspace import Document
from vyper.compiler import CompilerData
from vyper.exceptions import VyperException
from vyper.ast import nodes
from vyper_lsp.analyzer.BaseAnalyzer import Analyzer
from vyper_lsp.ast import AST
from vyper_lsp.utils import (
    get_expression_at_cursor,
    get_word_at_cursor,
    get_installed_vyper_version,
    get_internal_fn_name_at_cursor,
)
from lsprotocol.types import (
    CompletionItem,
    CompletionList,
    CompletionParams,
    SignatureHelpParams,
)
from pygls.server import LanguageServer

pattern = r"(.+) is deprecated\. Please use `(.+)` instead\."
compiled_pattern = re.compile(pattern)

min_vyper_version = Version("0.3.7")

# Available base types
UNSIGNED_INTEGER_TYPES = {f"uint{8*(i)}" for i in range(32, 0, -1)}
SIGNED_INTEGER_TYPES = {f"int{8*(i)}" for i in range(32, 0, -1)}
INTEGER_TYPES = UNSIGNED_INTEGER_TYPES | SIGNED_INTEGER_TYPES

BYTES_M_TYPES = {f"bytes{i}" for i in range(32, 0, -1)}
DECIMAL_TYPES = {"decimal"}

BASE_TYPES = {"bool", "address"} | INTEGER_TYPES | BYTES_M_TYPES | DECIMAL_TYPES

DECORATORS = ["payable", "nonpayable", "view", "pure", "external", "internal"]

logger = logging.getLogger("vyper-lsp")


class AstAnalyzer(Analyzer):
    def __init__(self, ast: AST) -> None:
        super().__init__()
        self.ast = ast
        if get_installed_vyper_version() < min_vyper_version:
            self.diagnostics_enabled = False
        else:
            self.diagnostics_enabled = True

    def _range_from_exception(self, node: VyperException) -> Range:
        return Range(
            start=Position(line=node.lineno - 1, character=node.col_offset),
            end=Position(line=node.end_lineno - 1, character=node.end_col_offset),
        )

    def _diagnostic_from_exception(self, node: VyperException) -> Diagnostic:
        return Diagnostic(
            range=self._range_from_exception(node),
            message=str(node),
            severity=1,
        )

    def signature_help(
        self, doc: Document, params: SignatureHelpParams
    ) -> SignatureHelp:
        current_line = doc.lines[params.position.line]
        expression = get_expression_at_cursor(
            current_line, params.position.character - 1
        )
        fn_name = get_internal_fn_name_at_cursor(
            current_line, params.position.character - 1
        )

        if expression.startswith("self."):
            node = self.ast.find_function_declaration_node_for_name(fn_name)
            if node:
                fn_name = node.name
                arg_str = ", ".join(
                    [f"{arg.arg}: {arg.annotation.id}" for arg in node.args.args]
                )
                fn_label = f"{fn_name}({arg_str})"
                parameters = []
                if node.returns:
                    line = doc.lines[node.lineno - 1]
                    fn_label = line.removeprefix("def ").removesuffix(":\n")
                for arg in node.args.args:
                    start_index = fn_label.find(arg.arg)
                    end_index = start_index + len(arg.arg)
                    parameters.append(
                        ParameterInformation(
                            label=(start_index, end_index), documentation=None
                        )
                    )
                active_parameter = current_line.split("(")[-1].count(",")
                return SignatureHelp(
                    signatures=[
                        SignatureInformation(
                            label=fn_label,
                            parameters=parameters,
                            documentation=None,
                            active_parameter=active_parameter or 0,
                        )
                    ],
                    active_signature=0,
                )

    def get_completions_in_doc(
        self, document: Document, params: CompletionParams
    ) -> CompletionList:
        items = []
        current_line = document.lines[params.position.line].strip()
        custom_types = self.ast.get_user_defined_types()

        if params.context:
            if params.context.trigger_character == ".":
                # get element before the dot
                element = current_line.split(" ")[-1].split(".")[0]

                # internal functions and state variables
                if element == "self":
                    for fn in self.ast.get_internal_functions():
                        items.append(CompletionItem(label=fn))
                    # TODO: This should exclude constants and immutables
                    for var in self.ast.get_state_variables():
                        items.append(CompletionItem(label=var))
                else:
                    # TODO: This is currently only correct for enums
                    # For structs, we'll need to get the type of the variable
                    for attr in self.ast.get_attributes_for_symbol(element):
                        items.append(CompletionItem(label=attr))
                completions = CompletionList(is_incomplete=False, items=items)
                return completions
            elif params.context.trigger_character == "@":
                for dec in DECORATORS:
                    items.append(CompletionItem(label=dec))
                completions = CompletionList(is_incomplete=False, items=items)
                return completions
            elif params.context.trigger_character == ":":
                for typ in custom_types + list(BASE_TYPES):
                    items.append(CompletionItem(label=typ, insert_text=f" {typ}"))

                completions = CompletionList(is_incomplete=False, items=items)
                return completions
            else:
                if params.context.trigger_character == " ":
                    if current_line[-1] == ":":
                        for typ in custom_types + list(BASE_TYPES):
                            items.append(CompletionItem(label=typ))

                        completions = CompletionList(is_incomplete=False, items=items)
                        return completions
                return CompletionList(is_incomplete=False, items=[])

        else:
            return CompletionList(is_incomplete=False, items=[])

    def get_completions(
        self, ls: LanguageServer, params: CompletionParams
    ) -> CompletionList:
        document = ls.workspace.get_text_document(params.text_document.uri)
        return self.get_completions_in_doc(document, params)

    def _is_internal_fn(self, expression: str) -> bool:
        return expression.startswith("self.") and "(" in expression

    def _is_state_var(self, expression: str) -> bool:
        return expression.startswith("self.") and "(" not in expression

    def _format_arg(self, arg: nodes.arg) -> str:
        if arg.annotation is None:
            return arg.arg

        # Handle case when annotation is a subscript (e.g., List[int])
        if isinstance(arg.annotation, nodes.Subscript):
            annotation_base = arg.annotation.value.id  # e.g., 'List' in 'List[int]'

            # Check if the subscript's slice is a simple name
            if isinstance(arg.annotation.slice.value, nodes.Name):
                annotation_subscript = (
                    arg.annotation.slice.value.id
                )  # e.g., 'int' in 'List[int]'
            else:
                annotation_subscript = (
                    arg.annotation.slice.value.value
                )  # Handle other subscript types

            return f"{arg.arg}: {annotation_base}[{annotation_subscript}]"

        # Default case for simple annotations
        return f"{arg.arg}: {arg.annotation.id}"

    def _format_fn_signature(self, node: nodes.FunctionDef) -> str:
        fn_name = node.name
        arg_str = ", ".join([self._format_arg(arg) for arg in node.args.args])
        if node.returns:
            if isinstance(node.returns, nodes.Subscript):
                return_type_str = (
                    f"{node.returns.value.id}[{node.returns.slice.value.value}]"
                )
            else:
                return_type_str = node.returns.id
            return (
                f"(Internal Function) **{fn_name}**({arg_str}) -> **{return_type_str}**"
            )
        return f"(Internal Function) **{fn_name}**({arg_str})"

    def hover_info(self, document: Document, pos: Position) -> Optional[str]:
        if len(document.lines) < pos.line:
            return None
        og_line = document.lines[pos.line]
        word = get_word_at_cursor(og_line, pos.character)
        full_word = get_expression_at_cursor(og_line, pos.character)

        if self._is_internal_fn(full_word):
            node = self.ast.find_function_declaration_node_for_name(word)
            if node:
                return self._format_fn_signature(node)
        elif self._is_state_var(full_word):
            node = self.ast.find_state_variable_declaration_node_for_name(word)
            if node:
                variable_type = node.annotation.id
                return f"(State Variable) **{word}** : **{variable_type}**"
        elif word in self.ast.get_structs():
            node = self.ast.find_type_declaration_node_for_name(word)
            if node:
                return f"(Struct) **{word}**"
        elif word in self.ast.get_enums():
            node = self.ast.find_type_declaration_node_for_name(word)
            if node:
                return f"(Enum) **{word}**"
        elif word in self.ast.get_events():
            node = self.ast.find_type_declaration_node_for_name(word)
            if node:
                return f"(Event) **{word}**"
        elif word in self.ast.get_constants():
            node = self.ast.find_state_variable_declaration_node_for_name(word)
            if node:
                variable_type = node.annotation.id
                return f"(Constant) **{word}** : **{variable_type}**"
        else:
            return None

    def get_diagnostics(self, doc: Document) -> List[Diagnostic]:
        diagnostics = []

        if not self.diagnostics_enabled:
            return diagnostics

        replacements = {}
        warnings.simplefilter("always")
        with warnings.catch_warnings(record=True) as w:
            try:
                compiler_data = CompilerData(doc.source)
                compiler_data.vyper_module_folded
            except VyperException as e:
                if e.lineno is not None and e.col_offset is not None:
                    diagnostics.append(self._diagnostic_from_exception(e))
                else:
                    for a in e.annotations:
                        diagnostics.append(self._diagnostic_from_exception(a))
            for warning in w:
                match = compiled_pattern.match(str(warning.message))
                if not match:
                    continue
                deprecated = match.group(1)
                replacement = match.group(2)
                replacements[deprecated] = replacement

        # iterate over doc.lines and find all deprecated values
        # and create a warning for each one at the correct position
        for i, line in enumerate(doc.lines):
            for deprecated, replacement in replacements.items():
                if deprecated in line:
                    diagnostics.append(
                        Diagnostic(
                            range=Range(
                                start=Position(
                                    line=i, character=line.index(deprecated)
                                ),
                                end=Position(
                                    line=i,
                                    character=line.index(deprecated) + len(deprecated),
                                ),
                            ),
                            message=f"{deprecated} is deprecated. Please use {replacement} instead.",
                            severity=2,
                        )
                    )

        return diagnostics
